import asyncio
from dataclasses import dataclass
import zmq
import zmq.asyncio
import logging
import socket
import ssl
import typing
from tornado import ioloop
from tornado.web import Application
from tornado.httpserver import HTTPServer as TornadoHTTP1Server
from tornado.httpclient import AsyncHTTPClient, HTTPRequest

# from tornado_http2.server import Server as TornadoHTTP2Server 
from ..param import Parameterized
from ..param.parameters import (Integer, IPAddress, ClassSelector, Selector, TypedList, String)
from .constants import ZMQ_PROTOCOLS, CommonRPC, HTTPServerTypes, ResourceTypes, ServerMessage
from .utils import get_IP_from_interface, issubklass
from .dataklasses import HTTPResource, ServerSentEvent
from .utils import get_default_logger
from .serializers import JSONSerializer
from .database import ThingInformation
from .zmq_message_brokers import  AsyncZMQClient, MessageMappedZMQClientPool
from .handlers import RPCHandler, BaseHandler, EventHandler, ThingsHandler, StopHandler
from .schema_validators import BaseSchemaValidator, JsonSchemaValidator
from .events import Event
from .eventloop import EventLoop
from .config import global_config




@dataclass 
class InteractionAffordance:
    URL_path : str
    obj : Event # typing.Union[Property, Action, Event]
    http_methods : typing.Tuple[str, typing.Optional[str], typing.Optional[str]]
    handler : BaseHandler
    kwargs : dict 

    def __eq__(self, other : "InteractionAffordance") -> bool:
        return self.obj == other.obj 
    
    

class HTTPServer(Parameterized):
    """
    HTTP(s) server to route requests to ``Thing``.
    """
    
    things = TypedList(item_type=str, default=None, allow_None=True, 
                       doc="instance name of the things to be served by the HTTP server." ) # type: typing.List[str]
    port = Integer(default=8080, bounds=(1, 65535),  
                    doc="the port at which the server should be run" ) # type: int
    address = IPAddress(default='0.0.0.0', 
                    doc="IP address") # type: str
    # protocol_version = Selector(objects=[1, 1.1, 2], default=2, 
    #                 doc="for HTTP 2, SSL is mandatory. HTTP2 is recommended. \
    #                 When no SSL configurations are provided, defaults to 1.1" ) # type: float
    logger = ClassSelector(class_=logging.Logger, default=None, allow_None=True, 
                    doc="logging.Logger" ) # type: logging.Logger
    log_level = Selector(objects=[logging.DEBUG, logging.INFO, logging.ERROR, logging.WARN, 
                                logging.CRITICAL, logging.ERROR], 
                    default=logging.INFO, 
                    doc="""alternative to logger, this creates an internal logger with the specified log level 
                    along with a IO stream handler.""" ) # type: int
    serializer = ClassSelector(class_=JSONSerializer,  default=None, allow_None=True,
                    doc="""json serializer used by the server""" ) # type: JSONSerializer
    ssl_context = ClassSelector(class_=ssl.SSLContext, default=None, allow_None=True, 
                    doc="SSL context to provide encrypted communication") # type: typing.Optional[ssl.SSLContext]    
    certfile = String(default=None, allow_None=True, 
                    doc="""alternative to SSL context, provide certificate file & key file to allow the server to 
                        create a SSL context""") # type: str
    keyfile = String(default=None, allow_None=True, 
                    doc="""alternative to SSL context, provide certificate file & key file to allow the server to 
                        create a SSL context""") # type: str
    allowed_clients = TypedList(item_type=str,
                            doc="""Serves request and sets CORS only from these clients, other clients are rejected with 403. 
                                Unlike pure CORS, the server resource is not even executed if the client is not 
                                an allowed client. if None any client is served.""")
    host = String(default=None, allow_None=True, 
                doc="Host Server to subscribe to coordinate starting sequence of things & web GUI" ) # type: str
    # network_interface = String(default='Ethernet',  
    #                         doc="Currently there is no logic to detect the IP addresss (as externally visible) correctly, \
    #                         therefore please send the network interface name to retrieve the IP. If a DNS server is present, \
    #                         you may leave this field" ) # type: str
    request_handler = ClassSelector(default=RPCHandler, class_=RPCHandler, isinstance=False, 
                            doc="custom web request handler of your choice for property read-write & action execution" ) # type: typing.Union[BaseHandler, RPCHandler]
    event_handler = ClassSelector(default=EventHandler, class_=(EventHandler, BaseHandler), isinstance=False, 
                            doc="custom event handler of your choice for handling events") # type: typing.Union[BaseHandler, EventHandler]
    schema_validator = ClassSelector(class_=BaseSchemaValidator, default=JsonSchemaValidator, allow_None=True, isinstance=False,
                        doc="""Validator for JSON schema. If not supplied, a default JSON schema validator is created.""") # type: BaseSchemaValidator
    
   
    
    def __init__(self, things : typing.List[str], *, port : int = 8080, address : str = '0.0.0.0', 
                host : typing.Optional[str] = None, logger : typing.Optional[logging.Logger] = None, log_level : int = logging.INFO, 
                serializer : typing.Optional[JSONSerializer] = None, ssl_context : typing.Optional[ssl.SSLContext] = None, 
                schema_validator : typing.Optional[BaseSchemaValidator] = JsonSchemaValidator,
                certfile : str = None, keyfile : str = None, 
                # protocol_version : int = 1, network_interface : str = 'Ethernet', 
                allowed_clients : typing.Optional[typing.Union[str, typing.Iterable[str]]] = None,   
                **kwargs) -> None:
        """
        Parameters
        ----------
        things: List[str]
            instance name of the things to be served as a list.
        port: int, default 8080
            the port at which the server should be run
        address: str, default 0.0.0.0
            IP address
        logger: logging.Logger, optional
            logging.Logger instance
        log_level: int
            alternative to logger, this creates an internal logger with the specified log level along with a IO stream handler. 
        serializer: JSONSerializer, optional
            json serializer used by the server
        ssl_context: ssl.SSLContext
            SSL context to provide encrypted communication
        certfile: str
            alternative to SSL context, provide certificate file & key file to allow the server to create a SSL context 
        keyfile: str
            alternative to SSL context, provide certificate file & key file to allow the server to create a SSL context 
        allowed_clients: List[str] 
            serves request and sets CORS only from these clients, other clients are reject with 403. Unlike pure CORS
            feature, the server resource is not even executed if the client is not an allowed client.
        **kwargs:
            rpc_handler: RPCHandler | BaseHandler, optional
                custom web request handler of your choice for property read-write & action execution
            event_handler: EventHandler | BaseHandler, optional
                custom event handler of your choice for handling events
        """
        super().__init__(
            things=things,
            port=port, 
            address=address, 
            host=host,
            logger=logger, 
            log_level=log_level,
            serializer=serializer or JSONSerializer(), 
            # protocol_version=1, 
            schema_validator=schema_validator,
            certfile=certfile, 
            keyfile=keyfile,
            ssl_context=ssl_context,
            # network_interface='Ethernet',# network_interface,
            request_handler=kwargs.get('request_handler', RPCHandler),
            event_handler=kwargs.get('event_handler', EventHandler),
            allowed_clients=allowed_clients if allowed_clients is not None else []
        )
        self._type = HTTPServerTypes.THING_SERVER
        self._lost_things = dict() # see update_router_with_thing
        self._zmq_protocol = ZMQ_PROTOCOLS.IPC
        self._zmq_inproc_socket_context = None 
        self._zmq_inproc_event_context = None
        self._local_rules = dict() # type: typing.Dict[str, typing.List[InteractionAffordance]]
 
    @property
    def all_ok(self) -> bool:
        self._IP = f"{self.address}:{self.port}"
        if self.logger is None:
            self.logger = get_default_logger('{}|{}'.format(self.__class__.__name__, 
                                            f"{self.address}:{self.port}"), 
                                            self.log_level)
            
        if self._zmq_protocol == ZMQ_PROTOCOLS.INPROC and (self._zmq_inproc_socket_context is None or self._zmq_inproc_event_context is None):
            raise ValueError("Inproc socket context is not provided. Logic Error.")
            
        self.app = Application(handlers=[
            (r'/remote-objects', ThingsHandler, dict(request_handler=self.request_handler, 
                                                        event_handler=self.event_handler)),
            (r'/stop', StopHandler, dict(owner=self))
        ])
        
        self.zmq_client_pool = MessageMappedZMQClientPool(self.things, identity=self._IP, 
                                                    deserialize_server_messages=False, handshake=False,
                                                    http_serializer=self.serializer, 
                                                    context=self._zmq_inproc_socket_context,
                                                    protocol=self._zmq_protocol,
                                                    logger=self.logger
                                                )
        # print("client pool context", self.zmq_client_pool.context)
        event_loop = EventLoop.get_async_loop() # sets async loop for a non-possessing thread as well
        event_loop.call_soon(lambda : asyncio.create_task(self.update_router_with_things()))
        event_loop.call_soon(lambda : asyncio.create_task(self.subscribe_to_host()))
        event_loop.call_soon(lambda : asyncio.create_task(self.zmq_client_pool.poll()) )
        for client in self.zmq_client_pool:
            event_loop.call_soon(lambda : asyncio.create_task(client._handshake(timeout=60000)))

        self.tornado_event_loop = None 
        # set value based on what event loop we use, there is some difference 
        # between the asyncio event loop and the tornado event loop
        
        # if self.protocol_version == 2:
        #     raise NotImplementedError("Current HTTP2 is not implemented.")
        #     self.tornado_instance = TornadoHTTP2Server(self.app, ssl_options=self.ssl_context)
        # else:
        self.tornado_instance = TornadoHTTP1Server(self.app, ssl_options=self.ssl_context)
        return True
    

    async def subscribe_to_host(self):
        if self.host is None:
            return
        client = AsyncHTTPClient()
        for i in range(300): # try for five minutes
            try:
                res = await client.fetch(HTTPRequest(
                        url=f"{self.host}/subscribers",
                        method='POST',
                        body=JSONSerializer.dumps(dict(
                                hostname=socket.gethostname(),
                                IPAddress=get_IP_from_interface(self.network_interface), 
                                port=self.port, 
                                type=self._type,
                                https=self.ssl_context is not None 
                            )),
                        validate_cert=False,
                        headers={"content-type" : "application/json"}
                    ))
            except Exception as ex:
                self.logger.error(f"Could not subscribe to host {self.host}. error : {str(ex)}, error type : {type(ex)}.")
                if i >= 299:
                    raise ex from None
            else: 
                if res.code in [200, 201]:
                    self.logger.info(f"subsribed successfully to host {self.host}")
                    break
                elif i >= 299:
                    raise RuntimeError(f"could not subsribe to host {self.host}. response {JSONSerializer.loads(res.body)}")
            await asyncio.sleep(1)
        # we lose the client anyway so we close it. if we decide to reuse the client, changes needed
        client.close() 


    def listen(self) -> None:
        """
        Start HTTP server. This method is blocking, async event loops intending to schedule the HTTP server should instead use  
        the inner tornado instance's (``HTTPServer.tornado_instance``) listen() method. 
        """
        assert self.all_ok, 'HTTPServer all is not ok before starting' # Will always be True or cause some other exception   
        self.tornado_event_loop = ioloop.IOLoop.current()
        self.tornado_instance.listen(port=self.port, address=self.address)    
        self.logger.info(f'started webserver at {self._IP}, ready to receive requests.')
        self.tornado_event_loop.start()


    async def stop(self) -> None:
        """
        Stop the event loop & the HTTP server. This method is async and should be awaited, mostly within a request
        handler. The stop handler at the path '/stop' with POST request is already implemented.
        """
        self.tornado_instance.stop()
        self.zmq_client_pool.stop_polling()
        await self.tornado_instance.close_all_connections()
        if self.tornado_event_loop is not None:
            self.tornado_event_loop.stop()
        
       
    async def update_router_with_things(self) -> None:
        """
        updates HTTP router with paths from ``Thing`` (s)
        """
        await asyncio.gather(*[self.update_router_with_thing(client) for client in self.zmq_client_pool])

        
    async def update_router_with_thing(self, client : AsyncZMQClient):
        if client.instance_name in self._lost_things:
            # Just to avoid duplication of this call as we proceed at single client level and not message mapped level
            return 
        self._lost_things[client.instance_name] = client
        self.logger.info(f"attempting to update router with thing {client.instance_name}.")
        while True:
            try:
                await client.handshake_complete()
                resources = dict() # type: typing.Dict[str, HTTPResource]
                reply = (await client.async_execute(
                                instruction=CommonRPC.http_resource_read(client.instance_name), 
                                raise_client_side_exception=True
                            ))[ServerMessage.DATA]
                resources.update(reply)

                handlers = []
                for instruction, http_resource in resources.items():
                    if http_resource["what"] in [ResourceTypes.PROPERTY, ResourceTypes.ACTION]:
                        resource = HTTPResource(**http_resource)
                        handlers.append((resource.fullpath, self.request_handler, dict(
                                                                resource=resource,
                                                                validator=self.schema_validator(resource.argument_schema) if global_config.validate_schema_on_client and resource.argument_schema else None,
                                                                owner=self                                                     
                                                            )))
                    elif http_resource["what"] == ResourceTypes.EVENT:
                        resource = ServerSentEvent(**http_resource)
                        if resource.class_name in self._local_rules and any(ia.obj._obj_name == resource.obj_name for ia in self._local_rules[resource.class_name]):
                            for ia in self._local_rules[resource.class_name]:
                                if ia.obj._obj_name == resource.obj_name:
                                    handlers.append((f'/{client.instance_name}{ia.URL_path}', ia.handler, dict(resource=resource, validator=None, 
                                                                                owner=self, **ia.kwargs)))
                        else:
                            handlers.append((instruction, self.event_handler, dict(
                                                                resource=resource,
                                                                validator=None,
                                                                owner=self 
                                                            )))
                    """
                    for handler based tornado rule matcher, the Rule object has following
                    signature
                    
                    def __init__(
                        self,
                        matcher: "Matcher",
                        target: Any,
                        target_kwargs: Optional[Dict[str, Any]] = None,
                        name: Optional[str] = None,
                    ) -> None:

                    matcher - based on route
                    target - handler
                    target_kwargs - given to handler's initialize
                    name - ...

                    len == 2 tuple is route + handler
                    len == 3 tuple is route + handler + target kwargs
                
                    so we give (path, RPCHandler, {'resource' : HTTPResource})
                    
                    path is extracted from remote_method(URL_path='....')
                    RPCHandler is the base handler of this package for RPC purposes
                    resource goes into target kwargs as the HTTPResource generated by 
                        remote_method and RemoteParamater contains all the info given 
                        to make RPCHandler work
                    """
                self.app.wildcard_router.add_rules(handlers)
                self.logger.info(f"updated router with thing {client.instance_name}.")
                break
            except Exception as ex:
                print("error", ex)
                self.logger.error(f"error while trying to update router with thing - {str(ex)}. " +
                                  "Trying again in 5 seconds")
                await asyncio.sleep(5)
       
        try:
            reply = (await client.async_execute(
                        instruction=CommonRPC.object_info_read(client.instance_name), 
                        raise_client_side_exception=True
                    ))[ServerMessage.DATA]
            object_info = ThingInformation(**reply)
            object_info.http_server ="{}://{}:{}".format("https" if self.ssl_context is not None else "http", 
                                                socket.gethostname(), self.port)
    
            await client.async_execute(
                        instruction=CommonRPC.object_info_write(client.instance_name),
                        arguments=dict(value=object_info), 
                        raise_client_side_exception=True
                    )
        except Exception as ex:
            self.logger.error(f"error while trying to update thing with HTTP server details - {str(ex)}. " +
                                "Trying again in 5 seconds")
        self.zmq_client_pool.poller.register(client.socket, zmq.POLLIN)
        self._lost_things.pop(client.instance_name)


    def add_event(self, URL_path : str, event : Event, handler : typing.Optional[BaseHandler] = None, 
                **kwargs) -> None:
        """
        Add an event to be served by HTTP server

        Parameters
        ----------
        URL_path : str
            URL path to access the event
        event : Event
            Event to be served
        handler : BaseHandler, optional
            custom handler for the event
        kwargs : dict
            additional keyword arguments to be passed to the handler's __init__
        """
        if not isinstance(event, Event):
            raise TypeError("event should be of type Event")
        if not issubklass(handler, BaseHandler):
            raise TypeError("handler should be subclass of BaseHandler")
        if event.owner.__name__ not in self._local_rules:
            self._local_rules[event.owner.__name__] = []
        obj = InteractionAffordance(URL_path=URL_path, obj=event,  
                        http_methods=('GET',), handler=handler or self.event_handler,  
                        kwargs=kwargs)
        if obj not in self._local_rules[event.owner.__name__]:
            self._local_rules[event.owner.__name__].append(obj)
               
    
__all__ = [
    HTTPServer.__name__
]


import os
import threading
import typing
from sqlalchemy import create_engine, select, inspect as inspect_database
from sqlalchemy.ext import asyncio as asyncio_ext
from sqlalchemy.orm import sessionmaker
from sqlalchemy import Integer, String, JSON, LargeBinary
from sqlalchemy.orm import Mapped, mapped_column, DeclarativeBase, MappedAsDataclass
from sqlite3 import DatabaseError
from dataclasses import dataclass

from ..param import Parameterized
from .constants import JSONSerializable
from .config import global_config
from .utils import pep8_to_URL_path
from .serializers import PythonBuiltinJSONSerializer as JSONSerializer, BaseSerializer
from .property import Property



class ThingTableBase(DeclarativeBase):
    """SQLAlchemy base table for all thing related tables"""
    pass 
    
class SerializedProperty(MappedAsDataclass, ThingTableBase):
    """
    Property value is serialized before storing in database, therefore providing unified version for 
    SQLite and other relational tables
    """
    __tablename__ = "properties"

    instance_name  : Mapped[str] = mapped_column(String)
    name : Mapped[str] = mapped_column(String, primary_key=True)
    serialized_value : Mapped[bytes] = mapped_column(LargeBinary) 


class ThingInformation(MappedAsDataclass, ThingTableBase):
    __tablename__ = "things"

    instance_name  : Mapped[str] = mapped_column(String, primary_key=True)
    class_name     : Mapped[str] = mapped_column(String)
    script         : Mapped[str] = mapped_column(String)
    kwargs         : Mapped[JSONSerializable] = mapped_column(JSON)
    eventloop_instance_name : Mapped[str] = mapped_column(String)
    http_server    : Mapped[str] = mapped_column(String)
    level          : Mapped[int] = mapped_column(Integer)
    level_type     : Mapped[str] = mapped_column(String) # starting local to computer or global to system?

    def json(self):
        return {
            "instance_name" : self.instance_name,
            "class_name" : self.class_name,
            "script" : self.script,
            "kwargs" : self.kwargs,
            "eventloop_instance_name" : self.eventloop_instance_name,
            "http_server" : self.http_server,
            "level" : self.level, 
            "level_type" : self.level_type
        }
    

@dataclass 
class DeserializedProperty: # not part of database
    """
    Property with deserialized value
    """
    instance_name : str
    name : str 
    value : typing.Any



class BaseDB:
    """
    Implements configuration file reader for all irrespective sync or async DB operation 
    """

    def __init__(self, instance : Parameterized, serializer : typing.Optional[BaseSerializer] = None, 
                config_file : typing.Union[str, None] = None) -> None:
        self.thing_instance = instance
        self.instance_name = instance.instance_name
        self.serializer = serializer
        self.URL = self.create_URL(config_file)
        self._batch_call_context = {}


    @classmethod
    def load_conf(cls, config_file : str) -> typing.Dict[str, typing.Any]:
        """
        load configuration file using JSON serializer
        """
        if not config_file:
            conf = {}
        elif config_file.endswith('.json'):
            file = open(config_file, 'r')
            conf = JSONSerializer.load(file)
        else:
            raise ValueError("config files of extension - {} expected, given file name {}".format(["json"], config_file))
        return conf
            
    def create_URL(self, config_file : str) -> str:
        """
        auto chooses among the different supported databases based on config file and creates the DB URL 
        """
        if config_file is None:
            folder = self.get_temp_dir_for_class_name(self.thing_instance.__class__.__name__)
            if not os.path.exists(folder):
                os.makedirs(folder)
            return BaseDB.create_sqlite_URL(**dict(file=f'{folder}{os.sep}{self.instance_name}.db'))
        conf = BaseDB.load_conf(config_file)
        if conf.get('server', None):
            return BaseDB.create_postgres_URL(conf=conf)
        else:
            return BaseDB.create_sqlite_URL(conf=conf)
    
    @classmethod
    def get_temp_dir_for_class_name(self, class_name : str) -> str:
        """
        get temporary directory for database files
        """
        return f"{global_config.TEMP_DIR}{os.sep}databases{os.sep}{pep8_to_URL_path(class_name)}"
    
    @classmethod
    def create_postgres_URL(cls, conf : str = None, database : typing.Optional[str] = None, 
                use_dialect : typing.Optional[bool] = False) -> str:
        """
        create a postgres URL
        """
        server = conf.get('server', None) 
        database = conf.get('database', database)
        host = conf.get("host", 'localhost')
        port = conf.get("port", 5432)
        user = conf.get('user', 'postgres')
        password = conf.get('password', '')
        if use_dialect:
            dialect = conf.get('dialect', None)
            if dialect:
                return f"{server}+{dialect}://{user}:{password}@{host}:{port}/{database}"      
        return f"{server}://{user}:{password}@{host}:{port}/{database}"
     
    @classmethod
    def create_sqlite_URL(self, **conf : typing.Dict[str, JSONSerializable]) -> str:
        """
        create sqlite URL
        """
        in_memory = conf.get('in_memory', False)
        dialect = conf.get('dialect', 'pysqlite')
        if not in_memory:
            file = conf.get('file', f'{global_config.TEMP_DIR}{os.sep}databases{os.sep}default.db')
            return f"sqlite+{dialect}:///{file}"
        else: 
            return f"sqlite+{dialect}:///:memory:"

    @property
    def in_batch_call_context(self):
        return threading.get_ident() in self._batch_call_context


class BaseAsyncDB(BaseDB):
    """
    Base class for an async database engine, implements configuration file reader, 
    sqlalchemy engine & session creation. Set ``async_db_engine`` boolean flag to True in ``Thing`` class 
    to use this engine. Database operations are then scheduled in the event loop instead of blocking the current thread.
    Scheduling happens after properties are set/written.

    Parameters
    ----------
    database: str
        The database to open in the database server specified in config_file (see below)
    serializer: BaseSerializer
        The serializer to use for serializing and deserializing data (for example
        property serializing before writing to database). Will be the same as zmq_serializer supplied to ``Thing``.
    config_file: str
        absolute path to database server configuration file
    """
    
    def __init__(self, instance : Parameterized, 
                serializer : typing.Optional[BaseSerializer] = None, 
                config_file : typing.Union[str, None] = None) -> None:
        super().__init__(instance=instance, serializer=serializer, config_file=config_file)
        self.engine = asyncio_ext.create_async_engine(self.URL)
        self.async_session = sessionmaker(self.engine, expire_on_commit=True, 
                        class_=asyncio_ext.AsyncSession)
        ThingTableBase.metadata.create_all(self.engine)



class BaseSyncDB(BaseDB):
    """
    Base class for an synchronous (blocking) database engine, implements configuration file reader, sqlalchemy engine 
    & session creation. Default DB engine for ``Thing`` & called immediately after properties are set/written.

    Parameters
    ----------
    database: str
        The database to open in the database server specified in config_file (see below)
    serializer: BaseSerializer
        The serializer to use for serializing and deserializing data (for example
        property serializing into database for storage). Will be the same as
        zmq_serializer supplied to ``Thing``.
    config_file: str
        absolute path to database server configuration file
    """

    def __init__(self, instance : Parameterized, 
                serializer : typing.Optional[BaseSerializer] = None, 
                config_file : typing.Union[str, None] = None) -> None:
        super().__init__(instance=instance, serializer=serializer, config_file=config_file)
        self.engine = create_engine(self.URL)
        self.sync_session = sessionmaker(self.engine, expire_on_commit=True)
        ThingTableBase.metadata.create_all(self.engine)
        
        

class ThingDB(BaseSyncDB):
    """
    Database engine composed within ``Thing``, carries out database operations like storing object information, properties 
    etc. 

    Parameters
    ----------
    instance_name: str
        ``instance_name`` of the ``Thing``
    serializer: BaseSerializer
        serializer used by the ``Thing``. The serializer to use for serializing and deserializing data (for example 
        property serializing into database for storage).
    config_file: str
        configuration file of the database server
    """

    def fetch_own_info(self): # -> ThingInformation:
        """
        fetch ``Thing`` instance's own information (some useful metadata which helps the ``Thing`` run).
        
        Returns
        -------
        ``ThingInformation``
        """
        if not inspect_database(self.engine).has_table("things"):
            return
        with self.sync_session() as session:
            stmt = select(ThingInformation).filter_by(instance_name=self.instance_name)
            data = session.execute(stmt)
            data = data.scalars().all()
            if len(data) == 0:
                return None 
            elif len(data) == 1:
                return data[0]
            else:
                raise DatabaseError("Multiple things with same instance name found, either cleanup database/detach/make new")
            
    
    def get_property(self, property : typing.Union[str, Property], deserialized : bool = True) -> typing.Any:
        """
        fetch a single property.

        Parameters
        ----------
        property: str | Property
            string name or descriptor object
        deserialized: bool, default True
            deserialize the property if True

        Returns
        -------
        value: Any
            property value
        """
        with self.sync_session() as session:
            name = property if isinstance(property, str) else property.name
            stmt = select(SerializedProperty).filter_by(instance_name=self.instance_name, name=name)
            data = session.execute(stmt)
            prop = data.scalars().all() # type: typing.Sequence[SerializedProperty]
            if len(prop) == 0:
                raise DatabaseError(f"property {name} not found in database")
            elif len(prop) > 1:
                raise DatabaseError("multiple properties with same name found") # Impossible actually
            if not deserialized:
                return prop[0]
            return self.serializer.loads(prop[0].serialized_value)
            

    def set_property(self, property : typing.Union[str, Property], value : typing.Any) -> None:
        """
        change the value of an already existing property.

        Parameters
        ----------
        property: str | Property
            string name or descriptor object
        value: Any
            value of the property
        """
        if self.in_batch_call_context:
            self._batch_call_context[threading.get_ident()][property.name] = value
            return
        with self.sync_session() as session:
            name = property if isinstance(property, str) else property.name
            stmt = select(SerializedProperty).filter_by(instance_name=self.instance_name, 
                                                name=name)
            data = session.execute(stmt)
            prop = data.scalars().all()
            if len(prop) > 1:
                raise DatabaseError("multiple properties with same name found") # Impossible actually
            if len(prop) == 1:
                prop = prop[0]
                prop.serialized_value = self.serializer.dumps(value)
            else:
                prop = SerializedProperty(
                        instance_name=self.instance_name, 
                        name=name,
                        serialized_value=self.serializer.dumps(getattr(self.thing_instance, name))
                    )
                session.add(prop)
            session.commit()

        
    def get_properties(self, properties : typing.Dict[typing.Union[str, Property], typing.Any],
                                                        deserialized : bool = True) -> typing.Dict[str, typing.Any]:
        """
        get multiple properties at once.

        Parameters
        ----------
        properties: List[str | Property]
            string names or the descriptor of the properties as a list

        Returns
        -------
        value: Dict[str, Any]
            property names and values as items
        """
        with self.sync_session() as session:
            names = []
            for obj in properties.keys():
                names.append(obj if isinstance(obj, str) else obj.name)
            stmt = select(SerializedProperty).filter_by(instance_name=self.instance_name).filter(
                                    SerializedProperty.name.in_(names))
            data = session.execute(stmt)
            unserialized_props = data.scalars().all()
            props = dict()
            for prop in unserialized_props:
                props[prop.name] = prop.serialized_value if not deserialized else self.serializer.loads(prop.serialized_value)
            return props


    def set_properties(self, properties : typing.Dict[typing.Union[str, Property], typing.Any]) -> None:
        """
        change the values of already existing few properties at once

        Parameters
        ----------
        properties: Dict[str | Property, Any]
            string names or the descriptor of the property and any value as dictionary pairs 
        """
        if self.in_batch_call_context:
            for obj, value in properties.items():
                name = obj if isinstance(obj, str) else obj.name
                self._batch_call_context[threading.get_ident()][name] = value
            return
        with self.sync_session() as session:
            names = []
            for obj in properties.keys():
                names.append(obj if isinstance(obj, str) else obj.name)
            stmt = select(SerializedProperty).filter_by(instance_name=self.instance_name).filter(
                                    SerializedProperty.name.in_(names))
            data = session.execute(stmt)
            db_props = data.scalars().all()
            for obj, value in properties.items():
                name = obj if isinstance(obj, str) else obj.name
                db_prop = list(filter(lambda db_prop: db_prop.name == name, db_props)) # type: typing.List[SerializedProperty]
                if len(prop) > 1:
                    raise DatabaseError("multiple properties with same name found") # Impossible actually
                if len(db_prop) == 1:
                    db_prop = db_prop[0] # type: SerializedProperty
                    db_prop.serialized_value = self.serializer.dumps(value)
                else:
                    prop = SerializedProperty(
                        instance_name=self.instance_name, 
                        name=name,
                        serialized_value=self.serializer.dumps(value)
                    )
                    session.add(prop)
            session.commit()

            
    def get_all_properties(self, deserialized : bool = True) -> typing.Dict[str, typing.Any]:
        """
        read all properties of the ``Thing`` instance.

        Parameters
        ----------
        deserialized: bool, default True
            deserilize the properties if True
        """
        with self.sync_session() as session:
            stmt = select(SerializedProperty).filter_by(instance_name=self.instance_name)
            data = session.execute(stmt)
            existing_props = data.scalars().all() # type: typing.Sequence[SerializedProperty]
            if not deserialized:
                return existing_props
            props = dict()
            for prop in existing_props:
                props[prop.name] = self.serializer.loads(prop.serialized_value)
            return props
        
        
    def create_missing_properties(self, properties : typing.Dict[str, Property],
                            get_missing_property_names : bool = False) -> None:
        """
        create any and all missing properties of ``Thing`` instance
        in database.

        Parameters
        ----------
        properties: Dict[str, Property]
            descriptors of the properties

        Returns
        -------
        List[str]
            list of missing properties if get_missing_propertys is True
        """
        missing_props = []
        with self.sync_session() as session:
            existing_props = self.get_all_properties()
            for name, new_prop in properties.items():
                if name not in existing_props: 
                    prop = SerializedProperty(
                        instance_name=self.instance_name, 
                        name=new_prop.name, 
                        serialized_value=self.serializer.dumps(getattr(self.thing_instance, 
                                                                new_prop.name))
                    )
                    session.add(prop)
                    missing_props.append(name)
            session.commit()
        if get_missing_property_names:
            return missing_props
        
   

class batch_db_commit:
    """
    Context manager to write multiple properties to database at once. Useful for sequential sets/writes of multiple properties 
    which has db_commit or db_persist set to True, but only write their values to database at once. 
    """
    def __init__(self, db_engine : ThingDB) -> None:
        self.db_engine = db_engine

    def __enter__(self) -> None: 
        self.db_engine._context[threading.get_ident()] = dict()
        
    def __exit__(self, exc_type, exc_value, exc_tb) -> None:
        data = self.db_engine._context.pop(threading.get_ident(), dict()) # typing.Dict[str, typing.Any]
        if exc_type is None:
            self.db_engine.set_properties(data)
            return 
        for name, value in data.items():
            try:
                self.db_engine.set_property(name, value)
            except Exception as ex:
                pass


  
__all__ = [
    BaseAsyncDB.__name__,
    BaseSyncDB.__name__,
    ThingDB.__name__,
    batch_db_commit.__name__
]


import typing, inspect
from dataclasses import dataclass, field


from .constants import JSON, JSONSerializable
from .utils import getattr_without_descriptor_read
from .dataklasses import ActionInfoValidator
from .events import Event
from .properties import *
from .property import Property
from .thing import Thing
from .state_machine import StateMachine



@dataclass
class Schema:
    """
    Base dataclass for all WoT schema; Implements a custom asdict method which replaces dataclasses' asdict 
    utility function
    """

    skip_keys = [] # override this to skip some dataclass attributes in the schema

    replacement_keys = {
        'context' : '@context',
        'htv_methodName' : 'htv:methodName'
    }

    def asdict(self):
        """dataclass fields as dictionary skip_keys and replacement_keys accounted"""
        schema = dict()
        for field, value in self.__dataclass_fields__.items():    
            if getattr(self, field, NotImplemented) is NotImplemented or field in self.skip_keys:
                continue
            if field in self.replacement_keys: 
                schema[self.replacement_keys[field]] = getattr(self, field)
            else: 
                schema[field] = getattr(self, field)
        return schema
    
    @classmethod
    def format_doc(cls, doc : str):
        """strip tabs, newlines, whitespaces etc."""
        doc_as_list = doc.split('\n')
        final_doc = []
        for index, line in enumerate(doc_as_list):
            line = line.lstrip('\n').rstrip('\n')
            line = line.lstrip('\t').rstrip('\t')
            line = line.lstrip('\n').rstrip('\n')
            line = line.lstrip().rstrip()   
            if index > 0:
                line = ' ' + line # add space to left in case of new line            
            final_doc.append(line)
        final_doc = ''.join(final_doc)
        final_doc = final_doc.lstrip().rstrip()
        return final_doc
    


class JSONSchema:
    """type restrictor converting python types to JSON schema types"""

    _allowed_types = ('string', 'object', 'array',  'number', 'integer', 'boolean', None)

    _replacements = {
        int : 'integer',
        float : 'number',
        str : 'string',
        bool : 'boolean',
        dict : 'object',
        list : 'array',
        tuple : 'array',
        type(None) : 'null',
        Exception : {
            "type": "object",
            "properties": {
                "message": {"type": "string"},
                "type": {"type": "string"},
                "traceback": {"type": "array", "items": {"type": "string"}},
                "notes": {"type": ["string", "null"]}
            },
            "required": ["message", "type", "traceback"]
        }
    }

    _schemas = {

    }

    @classmethod
    def is_allowed_type(cls, type : typing.Any) -> bool: 
        if type in JSONSchema._replacements.keys():
            return True 
        return False 
    
    @classmethod
    def is_supported(cls, typ: typing.Any) -> bool:
        """"""
        if typ in JSONSchema._schemas.keys():
            return True 
        return False 
    
    @classmethod
    def get_type(cls, typ : typing.Any) -> str:
        if not JSONSchema.is_allowed_type(typ):
            raise TypeError(f"Object for wot-td has invalid type for JSON conversion. Given type - {type(typ)}. " +
                                "Use JSONSchema.register_replacements on hololinked.wot.td.JSONSchema object to recognise the type.")
        return JSONSchema._replacements[typ]
    
    @classmethod
    def register_type_replacement(self, type : typing.Any, json_schema_type : str, 
                                schema : typing.Optional[typing.Dict[str, JSONSerializable]] = None) -> None:
        """
        specify a python type as a JSON type.
        schema only supported for array and objects. 
        """
        if json_schema_type in JSONSchema._allowed_types:
            JSONSchema._replacements[type] = json_schema_type
            if schema is not None:
                if json_schema_type not in ('array', 'object'):
                    raise ValueError(f"schemas support only for array and object JSON schema types, your specified type - {type}.")
                JSONSchema._schemas[type] = schema
        else:
            raise TypeError(f"json schema replacement type must be one of allowed type - 'string', 'object', 'array', 'string', " +
                                f"'number', 'integer', 'boolean', 'null'. Given value {json_schema_type}")

    @classmethod
    def get(cls, typ : typing.Any):
        """schema for array and objects only supported"""
        if not JSONSchema.is_supported(typ):
            raise ValueError(f"Schema for {typ} not provided. register one with JSONSchema.register_type_replacement()")
        return JSONSchema._schemas[typ]



@dataclass
class InteractionAffordance(Schema):
    """
    implements schema common to all interaction affordances. 
    concepts - https://www.w3.org/TR/wot-thing-description11/#interactionaffordance
    """
    title : str 
    titles : typing.Optional[typing.Dict[str, str]]
    description : str
    descriptions : typing.Optional[typing.Dict[str, str]] 
    forms : typing.List["Form"]
    # uri variables 

    def __init__(self):
        super().__init__()



@dataclass
class DataSchema(Schema):
    """
    implementes Dataschema attributes.
    https://www.w3.org/TR/wot-thing-description11/#sec-data-schema-vocabulary-definition
    """
    title : str 
    titles : typing.Optional[typing.Dict[str, str]]
    description : str
    descriptions : typing.Optional[typing.Dict[str, str]] 
    const : bool
    default : typing.Optional[typing.Any] 
    readOnly : bool
    writeOnly : bool # write only are to be considered actions with no return value
    format : typing.Optional[str]
    unit : typing.Optional[str]
    type : str
    oneOf : typing.Optional[typing.List[typing.Dict[str, JSONSerializable]]]
    enum : typing.Optional[typing.List[typing.Any]]

    def __init__(self):
        super().__init__()

    def build(self, property : Property, owner : Thing, authority : str) -> None:
        """generates the schema"""
        self.title = property.label or property.name 
        if property.constant:
            self.const = property.constant 
        if property.readonly:
            self.readOnly = property.readonly
        if property.fget is None:
            self.default = property.default
        if property.doc:
            self.description = Schema.format_doc(property.doc)
        if property.metadata and property.metadata.get("unit", None) is not None:
            self.unit = property.metadata["unit"]
        if property.allow_None:
            if not hasattr(self, 'oneOf'):
                self.oneOf = []
            if hasattr(self, 'type'):
                self.oneOf.append(dict(type=self.type))
                del self.type
            if not any(types["type"] == None for types in self.oneOf):
                self.oneOf.append(dict(type="null"))



@dataclass
class PropertyAffordance(InteractionAffordance, DataSchema):
    """
    creates property affordance schema from ``property`` descriptor object 
    schema - https://www.w3.org/TR/wot-thing-description11/#propertyaffordance
    """
    observable : bool

    _custom_schema_generators = dict()

    def __init__(self):
        super().__init__()

    def build(self, property : Property, owner : Thing, authority : str) -> None:
        """generates the schema"""
        DataSchema.build(self, property, owner, authority)

        self.forms = []
        for index, method in enumerate(property._remote_info.http_method):
            form = Form()
            # index is the order for http methods for (get, set, delete), generally (GET, PUT, DELETE)
            if (index == 1 and property.readonly) or index >= 2:
                continue # delete property is not a part of WoT, we also mostly never use it, so ignore.
            elif index == 0:
                form.op = 'readproperty'
            elif index == 1:
                form.op = 'writeproperty'
            form.href = f"{authority}{owner._full_URL_path_prefix}{property._remote_info.URL_path}"
            form.htv_methodName = method.upper()
            form.contentType = "application/json"
            self.forms.append(form.asdict())

        if property._observable:
            self.observable = property._observable
            form = Form()
            form.op = 'observeproperty'
            form.href = f"{authority}{owner._full_URL_path_prefix}{property._observable_event_descriptor.URL_path}"
            form.htv_methodName = "GET"
            form.subprotocol = "sse"
            form.contentType = "text/plain"
            self.forms.append(form.asdict())


    @classmethod
    def generate_schema(self, property : Property, owner : Thing, authority : str) -> typing.Dict[str, JSONSerializable]:
        if not isinstance(property, Property):
            raise TypeError(f"Property affordance schema can only be generated for Property. "
                            f"Given type {type(property)}")
        if isinstance(property, (String, Filename, Foldername, Path)):
            schema = StringSchema()
        elif isinstance(property, (Number, Integer)):
            schema = NumberSchema()
        elif isinstance(property, Boolean):
            schema = BooleanSchema()
        elif isinstance(property, (List, TypedList, Tuple, TupleSelector)):
            schema = ArraySchema()
        elif isinstance(property, Selector):
            schema = EnumSchema()
        elif isinstance(property, (TypedDict, TypedKeyMappingsDict)):
            schema = ObjectSchema()       
        elif isinstance(property, ClassSelector):
            schema = OneOfSchema()
        elif self._custom_schema_generators.get(property, NotImplemented) is not NotImplemented:
            schema = self._custom_schema_generators[property]()
        elif isinstance(property, Property) and property.model is not None:
            from .td_pydantic_extensions import GenerateJsonSchemaWithoutDefaultTitles, type_to_dataschema
            schema = PropertyAffordance()
            schema.build(property=property, owner=owner, authority=authority)
            data_schema = type_to_dataschema(property.model).model_dump(mode='json', exclude_none=True)
            final_schema = schema.asdict()
            if schema.oneOf: # allow_None = True
                final_schema['oneOf'].append(data_schema)
            else:
                final_schema.update(data_schema)
            return final_schema
        else:
            raise TypeError(f"WoT schema generator for this descriptor/property is not implemented. name {property.name} & type {type(property)}")     
        schema.build(property=property, owner=owner, authority=authority)
        return schema.asdict()
    
    @classmethod
    def register_descriptor(cls, descriptor : Property, schema_generator : "PropertyAffordance") -> None:
        if not isinstance(descriptor, Property):
            raise TypeError("custom schema generator can also be registered for Property." +
                            f" Given type {type(descriptor)}")
        if not isinstance(schema_generator, PropertyAffordance):
            raise TypeError("schema generator for Property must be subclass of PropertyAfforance. " +
                            f"Given type {type(schema_generator)}" )
        PropertyAffordance._custom_schema_generators[descriptor] = schema_generator


@dataclass
class BooleanSchema(PropertyAffordance):
    """
    boolean schema - https://www.w3.org/TR/wot-thing-description11/#booleanschema
    used by Boolean descriptor    
    """
    def __init__(self):
        super().__init__()

    def build(self, property: Property, owner: Thing, authority: str) -> None:
        """generates the schema"""
        self.type = 'boolean'
        PropertyAffordance.build(self, property, owner, authority)


@dataclass
class StringSchema(PropertyAffordance):
    """
    string schema - https://www.w3.org/TR/wot-thing-description11/#stringschema
    used by String, Filename, Foldername, Path descriptors
    """
    pattern : typing.Optional[str] 
    
    def __init__(self):
        super().__init__()
        
    def build(self, property: Property, owner: Thing, authority: str) -> None:
        """generates the schema"""
        self.type = 'string' 
        PropertyAffordance.build(self, property, owner, authority)
        if isinstance(property, String): 
            if property.regex is not None:
                self.pattern = property.regex


@dataclass
class NumberSchema(PropertyAffordance):
    """
    number schema - https://www.w3.org/TR/wot-thing-description11/#numberschema
    used by String, Filename, Foldername, Path descriptors
    """
    minimum : typing.Optional[typing.Union[int, float]]
    maximum : typing.Optional[typing.Union[int, float]]
    exclusiveMinimum : typing.Optional[bool] 
    exclusiveMaximum : typing.Optional[bool] 
    step : typing.Optional[typing.Union[int, float]]

    def __init__(self):
        super().__init__()
        
    def build(self, property: Property, owner: Thing, authority: str) -> None:
        """generates the schema"""
        if isinstance(property, Integer):
            self.type = 'integer'
        elif isinstance(property, Number): # dont change order - one is subclass of other
            self.type = 'number' 
        PropertyAffordance.build(self, property, owner, authority)
        if property.bounds is not None:      
            if isinstance(property.bounds[0], (int, float)): # i.e. value is not None which is allowed by param
                if not property.inclusive_bounds[0]:
                    self.exclusiveMinimum = property.bounds[0]
                else:
                    self.minimum = property.bounds[0]
            if isinstance(property.bounds[1], (int, float)):
                if not property.inclusive_bounds[1]:
                    self.exclusiveMaximum = property.bounds[1]
                else:
                    self.maximum = property.bounds[1]
        if property.step:
            self.multipleOf = property.step


@dataclass
class ArraySchema(PropertyAffordance):
    """
    array schema - https://www.w3.org/TR/wot-thing-description11/#arrayschema
    Used by List, Tuple, TypedList and TupleSelector
    """

    items : typing.Optional[typing.Dict[str, JSONSerializable]]
    minItems : typing.Optional[int]
    maxItems : typing.Optional[int]

    def __init__(self):
        super().__init__()
        
    def build(self, property: Property, owner: Thing, authority: str) -> None:
        """generates the schema"""
        self.type = 'array'
        PropertyAffordance.build(self, property, owner, authority)
        self.items = []
        if isinstance(property, (List, Tuple, TypedList)) and property.item_type is not None:
            if property.bounds:
                if property.bounds[0]:
                    self.minItems = property.bounds[0]
                if property.bounds[1]:
                    self.maxItems = property.bounds[1]
            if isinstance(property.item_type, (list, tuple)):
                for typ in property.item_type:
                    self.items.append(dict(type=JSONSchema.get_type(typ)))
            elif property.item_type is not None: 
                self.items.append(dict(type=JSONSchema.get_type(property.item_type)))
        elif isinstance(property, TupleSelector):
            objects = list(property.objects)
            for obj in objects:
                if any(types["type"] == JSONSchema._replacements.get(type(obj), None) for types in self.items):
                    continue 
                self.items.append(dict(type=JSONSchema.get_type(type(obj))))
        if len(self.items) == 0:
            del self.items
        elif len(self.items) > 1:
            self.items = dict(oneOf=self.items)
            

@dataclass
class ObjectSchema(PropertyAffordance):
    """
    object schema - https://www.w3.org/TR/wot-thing-description11/#objectschema
    Used by TypedDict
    """
    properties : typing.Optional[typing.Dict[str, JSONSerializable]]
    required : typing.Optional[typing.List[str]]

    def __init__(self):
        super().__init__()
        
    def build(self, property: Property, owner: Thing, authority: str) -> None:
        """generates the schema"""
        PropertyAffordance.build(self, property, owner, authority)
        properties = None 
        required = None 
        if hasattr(property, 'json_schema'):
            # Code will not reach here for now as have not implemented schema for typed dictionaries. 
            properties = property.json_schema["properties"]
            if property.json_schema.get("required", NotImplemented) is not NotImplemented:
                required = property.json_schema["required"] 
        if not property.allow_None:
            self.type = 'object'
            if properties:
                self.properties = properties
            if required:
                self.required = required
        else:
            schema = dict(type='object')
            if properties:
                schema['properties'] = properties
            if required:
                schema['required'] = required
            self.oneOf.append(schema)


@dataclass
class OneOfSchema(PropertyAffordance):
    """
    custom schema to deal with ClassSelector to fill oneOf field correctly
    https://www.w3.org/TR/wot-thing-description11/#dataschema
    """
    properties : typing.Optional[typing.Dict[str, JSONSerializable]]
    required : typing.Optional[typing.List[str]]
    items : typing.Optional[typing.Dict[str, JSONSerializable]]
    minItems : typing.Optional[int]
    maxItems : typing.Optional[int]
    # ClassSelector can technically have a JSON serializable as a class_

    def __init__(self):
        super().__init__()

    def build(self, property: Property, owner: Thing, authority: str) -> None:
        """generates the schema"""
        self.oneOf = []
        if isinstance(property, ClassSelector):
            if not property.isinstance:
                raise NotImplementedError("WoT TD for ClassSelector with isinstance set to True is not supported yet. "  +
                                          "Consider user this property in a different way.")
            if isinstance(property.class_, (list, tuple)):
                objects = list(property.class_)
            else:
                objects = [property.class_]
        elif isinstance(property, Selector):
            objects = list(property.objects)
        else:
            raise TypeError(f"EnumSchema and OneOfSchema supported only for Selector and ClassSelector. Given Type - {property}")
        for obj in objects:
            if any(types["type"] == JSONSchema._replacements.get(type(obj), None) for types in self.oneOf):
                continue 
            if isinstance(property, ClassSelector):
                if not JSONSchema.is_allowed_type(obj):
                    raise TypeError(f"Object for wot-td has invalid type for JSON conversion. Given type - {obj}. " +
                                "Use JSONSchema.register_replacements on hololinked.wot.td.JSONSchema object to recognise the type.")
                subschema = dict(type=JSONSchema.get_type(obj))
                if JSONSchema.is_supported(obj):
                    subschema.update(JSONSchema.get(obj))
                self.oneOf.append(subschema)
            elif isinstance(property, Selector):
                if JSONSchema.get_type(type(obj)) == "null":
                    continue
                self.oneOf.append(dict(type=JSONSchema.get_type(type(obj))))
        PropertyAffordance.build(self, property, owner, authority)
        self.cleanup()

    def cleanup(self):
        if len(self.oneOf) == 1:
            oneOf = self.oneOf[0]
            self.type = oneOf["type"]
            if oneOf["type"] == 'object':
                if oneOf.get("properties", NotImplemented) is not NotImplemented:
                    self.properties = oneOf["properties"]
                if oneOf.get("required", NotImplemented) is not NotImplemented:
                    self.required = oneOf["required"]
            elif oneOf["type"] == 'array':
                if oneOf.get("items", NotImplemented) is not NotImplemented:
                    self.items = oneOf["items"]
                if oneOf.get("maxItems", NotImplemented) is not NotImplemented:
                    self.minItems = oneOf["minItems"]
                if oneOf.get("maxItems", NotImplemented) is not NotImplemented:
                    self.maxItems = oneOf["maxItems"]
            del self.oneOf


@dataclass
class EnumSchema(OneOfSchema):
    """
    custom schema to fill enum field of property affordance correctly
    https://www.w3.org/TR/wot-thing-description11/#dataschema
    """ 
    def __init__(self):
        super().__init__()
        
    def build(self, property: Property, owner: Thing, authority: str) -> None:
        """generates the schema"""
        assert isinstance(property, Selector), f"EnumSchema compatible property is only Selector, not {property.__class__}"
        self.enum = list(property.objects)
        OneOfSchema.build(self, property, owner, authority)


@dataclass
class Link(Schema):
    href : str
    anchor : typing.Optional[str]  
    type : typing.Optional[str] = field(default='application/json')
    rel : typing.Optional[str] = field(default='next')

    def __init__(self):
        super().__init__()
    
    def build(self, resource : Thing, owner : Thing, authority : str) -> None:
        self.href = f"{authority}{resource._full_URL_path_prefix}/resources/wot-td"
        self.anchor = f"{authority}{owner._full_URL_path_prefix}"



@dataclass
class ExpectedResponse(Schema):
    """
    Form property. 
    schema - https://www.w3.org/TR/wot-thing-description11/#expectedresponse
    """
    contentType : str

    def __init__(self):
        super().__init__()


@dataclass
class AdditionalExpectedResponse(Schema):
    """
    Form field for additional responses which are different from the usual response.
    schema - https://www.w3.org/TR/wot-thing-description11/#additionalexpectedresponse
    """
    success : bool = field(default=False)
    contentType : str = field(default='application/json')
    schema : typing.Optional[JSON] = field(default='exception')

    def __init__(self):
        super().__init__()

   
@dataclass
class Form(Schema):
    """
    Form hypermedia.
    schema - https://www.w3.org/TR/wot-thing-description11/#form
    """
    href : str 
    op : str 
    htv_methodName : str 
    contentType : typing.Optional[str]
    additionalResponses : typing.Optional[typing.List[AdditionalExpectedResponse]]
    contentEncoding : typing.Optional[str]
    security : typing.Optional[str]
    scopes : typing.Optional[str]
    response : typing.Optional[ExpectedResponse]
    subprotocol : typing.Optional[str]
    
    def __init__(self):
        super().__init__()

    
@dataclass
class ActionAffordance(InteractionAffordance):
    """
    creates action affordance schema from actions (or methods).
    schema - https://www.w3.org/TR/wot-thing-description11/#actionaffordance
    """
    input : typing.Dict[str, JSONSerializable]
    output : typing.Dict[str, JSONSerializable]
    safe : bool
    idempotent : bool 
    synchronous : bool 

    def __init__(self):
        super(InteractionAffordance, self).__init__()
    
    def build(self, action : typing.Callable, owner : Thing, authority : str) -> None:
        assert isinstance(action._remote_info, ActionInfoValidator)
        if action._remote_info.argument_schema: 
            self.input = action._remote_info.argument_schema 
        if action._remote_info.return_value_schema: 
            self.output = action._remote_info.return_value_schema 
        self.title = action.__name__
        if action.__doc__:
            self.description = self.format_doc(action.__doc__)
        if not (hasattr(owner, 'state_machine') and owner.state_machine is not None and 
                owner.state_machine.has_object(action._remote_info.obj)) and action._remote_info.idempotent:
            self.idempotent = action._remote_info.idempotent
        if action._remote_info.synchronous:
            self.synchronous = action._remote_info.synchronous
        if action._remote_info.safe:
            self.safe = action._remote_info.safe 
        self.forms = []
        for method in action._remote_info.http_method:
            form = Form()
            form.op = 'invokeaction'
            form.href = f'{authority}{owner._full_URL_path_prefix}{action._remote_info.URL_path}'
            form.htv_methodName = method.upper()
            form.contentType = 'application/json'
            # form.additionalResponses = [AdditionalExpectedResponse().asdict()]
            self.forms.append(form.asdict())

    @classmethod
    def generate_schema(cls, action : typing.Callable, owner : Thing, authority : str) -> typing.Dict[str, JSONSerializable]:
        schema = ActionAffordance()
        schema.build(action=action, owner=owner, authority=authority) 
        return schema.asdict()
    

@dataclass
class EventAffordance(InteractionAffordance):
    """
    creates event affordance schema from events.
    schema - https://www.w3.org/TR/wot-thing-description11/#eventaffordance
    """
    subscription : str
    data : typing.Dict[str, JSONSerializable]
    
    def __init__(self):
        super().__init__()
    
    def build(self, event : Event, owner : Thing, authority : str) -> None:
        self.title = event.label or event._obj_name 
        if event.doc:
            self.description = self.format_doc(event.doc)
        if event.schema:
            self.data = event.schema

        form = Form()
        form.op = "subscribeevent"
        form.href = f"{authority}{owner._full_URL_path_prefix}{event.URL_path}"
        form.htv_methodName = "GET"
        form.contentType = "text/plain"
        form.subprotocol = "sse"
        self.forms = [form.asdict()]

    @classmethod
    def generate_schema(cls, event : Event, owner : Thing, authority : str) -> typing.Dict[str, JSONSerializable]:
        schema = EventAffordance()
        schema.build(event=event, owner=owner, authority=authority)
        return schema.asdict()


@dataclass
class VersionInfo:
    """
    create version info.
    schema - https://www.w3.org/TR/wot-thing-description11/#versioninfo
    """
    instance : str 
    model : str


@dataclass
class SecurityScheme(Schema):
    """
    create security scheme. 
    schema - https://www.w3.org/TR/wot-thing-description11/#sec-security-vocabulary-definition
    """
    scheme: str 
    description : str 
    descriptions : typing.Optional[typing.Dict[str, str]]
    proxy : typing.Optional[str]

    def __init__(self):
        super().__init__()

    def build(self, name : str, instance):
        self.scheme = 'nosec'
        self.description = 'currently no security scheme supported - use cookie auth directly on hololinked.server.HTTPServer object'
        return { name : self.asdict() }



@dataclass
class ThingDescription(Schema):
    """
    generate Thing Description schema of W3 Web of Things standard. 
    Refer standard - https://www.w3.org/TR/wot-thing-description11
    Refer schema - https://www.w3.org/TR/wot-thing-description11/#thing
    """
    context : typing.Union[typing.List[str], str, typing.Dict[str, str]] 
    type : typing.Optional[typing.Union[str, typing.List[str]]]
    id : str 
    title : str 
    description : str 
    version : typing.Optional[VersionInfo]
    created : typing.Optional[str] 
    modified : typing.Optional[str]
    support : typing.Optional[str] 
    base : typing.Optional[str] 
    properties : typing.List[PropertyAffordance]
    actions : typing.List[ActionAffordance]
    events : typing.List[EventAffordance]
    links : typing.Optional[typing.List[Link]] 
    forms : typing.Optional[typing.List[Form]]
    security : typing.Union[str, typing.List[str]]
    securityDefinitions : SecurityScheme
    schemaDefinitions : typing.Optional[typing.List[DataSchema]]
    
    skip_properties = ['expose', 'httpserver_resources', 'zmq_resources', 'gui_resources',
                    'events', 'thing_description', 'GUI', 'object_info' ]

    skip_actions = ['_set_properties', '_get_properties', '_add_property', '_get_properties_in_db', 
                    'push_events', 'stop_events', 'get_postman_collection', 'get_thing_description',
                    'get_our_temp_thing_description']

    # not the best code and logic, but works for now

    def __init__(self, instance : Thing, authority : typing.Optional[str] = None, 
                    allow_loose_schema : typing.Optional[bool] = False, ignore_errors : bool = False) -> None:
        super().__init__()
        self.instance = instance
        self.authority = authority
        self.allow_loose_schema = allow_loose_schema
        self.ignore_errors = ignore_errors

    def produce(self) -> typing.Dict[str, typing.Any]: 
        self.context = "https://www.w3.org/2022/wot/td/v1.1"
        self.id = f"{self.authority}/{self.instance.instance_name}"
        self.title = self.instance.__class__.__name__ 
        self.description = Schema.format_doc(self.instance.__doc__) if self.instance.__doc__ else "no class doc provided" 
        self.properties = dict()
        self.actions = dict()
        self.events = dict()
        self.forms = NotImplemented
        self.links = NotImplemented
        
        # self.schemaDefinitions = dict(exception=JSONSchema.get_type(Exception))

        self.add_interaction_affordances()
        self.add_top_level_forms()
        self.add_security_definitions()
       
        return self.asdict()
    

    def add_interaction_affordances(self):
        # properties and actions
        for resource in self.instance.instance_resources.values():
            try:
                if (resource.isproperty and resource.obj_name not in self.properties and 
                    resource.obj_name not in self.skip_properties and hasattr(resource.obj, "_remote_info") and 
                    resource.obj._remote_info is not None): 
                    if (resource.obj_name == 'state' and (not hasattr(self.instance, 'state_machine') or 
                                not isinstance(self.instance.state_machine, StateMachine))):
                        continue
                    self.properties[resource.obj_name] = PropertyAffordance.generate_schema(resource.obj, 
                                                                            self.instance, self.authority) 
                
                elif (resource.isaction and resource.obj_name not in self.actions and 
                    resource.obj_name not in self.skip_actions and hasattr(resource.obj, '_remote_info')):
                   
                    self.actions[resource.obj_name] = ActionAffordance.generate_schema(resource.obj, 
                                                                                self.instance, self.authority)
            except Exception as ex:
                if not self.ignore_errors:
                    raise ex from None
                self.instance.logger.error(f"Error while generating schema for {resource.obj_name} - {ex}")
        # Events
        for name, resource in inspect._getmembers(self.instance, lambda o : isinstance(o, Event),
                                                getattr_without_descriptor_read):
            if not isinstance(resource, Event):
                continue
            if '/change-event' in resource.URL_path:
                continue
            try:
                self.events[name] = EventAffordance.generate_schema(resource, self.instance, self.authority)
            except Exception as ex:
                if not self.ignore_errors:
                    raise ex from None
                self.instance.logger.error(f"Error while generating schema for {resource.obj_name} - {ex}")
        # for name, resource in inspect._getmembers(self.instance, lambda o : isinstance(o, Thing), getattr_without_descriptor_read):
        #     if resource is self.instance or isinstance(resource, EventLoop):
        #         continue
        #     if self.links is None:
        #         self.links = []
        #     link = Link()
        #     link.build(resource, self.instance, self.authority)
        #     self.links.append(link.asdict())
    

    def add_top_level_forms(self):

        self.forms = []

        properties_end_point = f"{self.authority}{self.instance._full_URL_path_prefix}/properties"

        readallproperties = Form()
        readallproperties.href = properties_end_point
        readallproperties.op = "readallproperties"
        readallproperties.htv_methodName = "GET"
        readallproperties.contentType = "application/json"
        # readallproperties.additionalResponses = [AdditionalExpectedResponse().asdict()]
        self.forms.append(readallproperties.asdict())
        
        writeallproperties = Form() 
        writeallproperties.href = properties_end_point
        writeallproperties.op = "writeallproperties"   
        writeallproperties.htv_methodName = "PUT"
        writeallproperties.contentType = "application/json" 
        # writeallproperties.additionalResponses = [AdditionalExpectedResponse().asdict()]
        self.forms.append(writeallproperties.asdict())

        readmultipleproperties = Form()
        readmultipleproperties.href = properties_end_point
        readmultipleproperties.op = "readmultipleproperties"
        readmultipleproperties.htv_methodName = "GET"
        readmultipleproperties.contentType = "application/json"
        # readmultipleproperties.additionalResponses = [AdditionalExpectedResponse().asdict()]
        self.forms.append(readmultipleproperties.asdict())

        writemultipleproperties = Form() 
        writemultipleproperties.href = properties_end_point
        writemultipleproperties.op = "writemultipleproperties"   
        writemultipleproperties.htv_methodName = "PATCH"
        writemultipleproperties.contentType = "application/json"
        # writemultipleproperties.additionalResponses = [AdditionalExpectedResponse().asdict()]
        self.forms.append(writemultipleproperties.asdict())
        
    def add_security_definitions(self):
        self.security = 'unimplemented'
        self.securityDefinitions = SecurityScheme().build('unimplemented', self.instance)



__all__ = [
    ThingDescription.__name__,
    JSONSchema.__name__
]


import logging, threading, time
import unittest
from hololinked.client import ObjectProxy
from hololinked.server import Thing, action, Event
from hololinked.server.properties import Number
try:
    from .utils import TestCase, TestRunner
    from .things import TestThing
except ImportError:
    from utils import TestCase, TestRunner
    from things import start_thing_forked 



class TestThing(Thing):

    total_number_of_events = Number(default=1, bounds=(1, None),                            
                            doc="Total number of events pushed")
    
    test_event = Event(friendly_name="test-event", doc="A test event", 
                URL_path='/test-event') 
    
    @action()
    def push_events(self):
        threading.Thread(target=self._push_worker).start()

    def _push_worker(self):
        for i in range(100):
            self.test_event.push('test data')
            time.sleep(0.01)



class TestEvent(TestCase):
    
    @classmethod
    def setUpClass(self):
        print("test event")
        self.thing_cls = TestThing
        start_thing_forked(self.thing_cls, instance_name='test-event',
                                    log_level=logging.WARN)   
        self.thing_client = ObjectProxy('test-event') # type: TestThing

    @classmethod
    def tearDownClass(self):
        print("tear down test event")
        self.thing_client.exit()
    
    
    def test_1_event(self):
        attempts = 100 
        self.thing_client.total_number_of_events = attempts

        results = []
        def cb(value):
            results.append(value)

        self.thing_client.test_event.subscribe(cb)
        time.sleep(3)
        # Calm down for event publisher to connect fully as there is no handshake for events
        self.thing_client.push_events()

        for i in range(attempts):
            if len(results) == attempts:
                break
            time.sleep(0.1)

        self.assertEqual(len(results), attempts)
        self.assertEqual(results, ['test data']*attempts)
        self.thing_client.test_event.unsubscribe(cb)



if __name__ == '__main__':
    unittest.main(testRunner=TestRunner())


import sys 
import os
import warnings
import subprocess
import asyncio
import importlib
import typing 
import threading
import logging
import tracemalloc
from uuid import uuid4

from .constants import HTTP_METHODS
from .utils import format_exception_as_json
from .config import global_config
from .zmq_message_brokers import ServerTypes 
from .exceptions import *
from .thing import Thing, ThingMeta
from .property import Property
from .properties import ClassSelector, TypedList, List, Boolean, TypedDict
from .action import action as remote_method
from .logger import ListHandler


if global_config.TRACE_MALLOC:
    tracemalloc.start()


def set_event_loop_policy():
    if sys.platform.lower().startswith('win'):
        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

    if global_config.USE_UVLOOP:
        if sys.platform.lower() in ['linux', 'darwin', 'linux2']:
            import uvloop
            asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
        else:
            warnings.warn("uvloop not supported for windows, using default windows selector loop.", RuntimeWarning)

set_event_loop_policy()


class Consumer:
    """
    Container class for Thing to pass to eventloop for multiprocessing applications in case 
    of rare needs. 
    """
    object_cls = ClassSelector(default=None, allow_None=True, class_=Thing, isinstance=False,
                            remote=False)
    args = List(default=None, allow_None=True, accept_tuple=True, remote=False)
    kwargs = TypedDict(default=None, allow_None=True, key_type=str, remote=False)
   
    def __init__(self, object_cls : typing.Type[Thing], args : typing.Tuple = tuple(), **kwargs) -> None:
        self.object_cls = object_cls
        self.args = args 
        self.kwargs = kwargs



RemoteObject = Thing # reading convenience

class EventLoop(RemoteObject):
    """
    The EventLoop class implements a infinite loop where zmq ROUTER sockets listen for messages. Each consumer of the 
    event loop (an instance of Thing) listen on their own ROUTER socket and execute methods or allow read and write
    of attributes upon receiving instructions. Socket listening is implemented in an async (asyncio) fashion. 
    """
    server_type = ServerTypes.EVENTLOOP

    expose = Boolean(default=True, remote=False,
                     doc="""set to False to use the object locally to avoid alloting network resources 
                        of your computer for this object""")

    things = TypedList(item_type=(Thing, Consumer), bounds=(0,100), allow_None=True, default=None,
                        doc="list of Things which are being executed", remote=False) #type: typing.List[Thing]
    
    threaded = Boolean(default=False, remote=False, 
                        doc="set True to run each thing in its own thread")
  

    def __init__(self, *, 
                instance_name : str, 
                things : typing.Union[Thing, Consumer, typing.List[typing.Union[Thing, Consumer]]] = list(), # type: ignore - requires covariant types
                log_level : int = logging.INFO, 
                **kwargs
            ) -> None:
        """
        Parameters
        ----------
        instance_name: str
            instance name of the event loop
        things: List[Thing]
            things to be run/served
        log_level: int
            log level of the event loop logger
        """
        super().__init__(instance_name=instance_name, things=things, log_level=log_level, **kwargs)
        things = [] # type: typing.List[Thing]
        if self.expose:
            things.append(self) 
        if self.things is not None:
            for consumer in self.things:
                if isinstance(consumer, Thing):
                    things.append(consumer)
                    consumer.object_info.eventloop_instance_name = self.instance_name
                elif isinstance(consumer, Consumer):
                    instance = consumer.object_cls(*consumer.args, **consumer.kwargs, 
                                            eventloop_name=self.instance_name)
                    things.append(instance) 
        self.things = things # re-assign the instantiated objects as well
        self.uninstantiated_things = dict()
        self._message_listener_methods = []

    def __post_init__(self):
        super().__post_init__()
        self.logger.info("Event loop with name '{}' can be started using EventLoop.run().".format(self.instance_name))   


    # example of overloading
    @remote_method()
    def exit(self):
        """
        Stops the event loop and all its things. Generally, this leads
        to exiting the program unless some code follows the ``run()`` method.  
        """
        for thing in self.things:
            thing.exit()
        raise BreakAllLoops
    

    uninstantiated_things = TypedDict(default=None, allow_None=True, key_type=str,
                        item_type=(Consumer, str), URL_path='/things/uninstantiated')
    
    
    @classmethod
    def _import_thing(cls, file_name : str, object_name : str):
        """
        import a thing specified by ``object_name`` from its 
        script or module. 

        Parameters
        ----------
        file_name : str
            file or module path 
        object_name : str
            name of ``Thing`` class to be imported
        """
        module_name = file_name.split(os.sep)[-1]
        spec = importlib.util.spec_from_file_location(module_name, file_name)
        if spec is not None:
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)
        else:     
            module = importlib.import_module(module_name, file_name.split(os.sep)[0])
        consumer = getattr(module, object_name) 
        if issubclass(consumer, Thing):
            return consumer 
        else:
            raise ValueError(f"object name {object_name} in {file_name} not a subclass of Thing.", 
                            f" Only subclasses are accepted (not even instances). Given object : {consumer}")
        

    @remote_method(URL_path='/things', http_method=HTTP_METHODS.POST)
    def import_thing(self, file_name : str, object_name : str):
        """
        import thing from the specified path and return the default 
        properties to be supplied to instantiate the object. 
        """
        consumer = self._import_thing(file_name, object_name) # type: ThingMeta
        id = uuid4()
        self.uninstantiated_things[id] = consumer
        return id
           

    @remote_method(URL_path='/things/instantiate', 
                http_method=HTTP_METHODS.POST) # remember to pass schema with mandatory instance name
    def instantiate(self, id : str, kwargs : typing.Dict = {}):      
        """
        Instantiate the thing that was imported with given arguments 
        and add to the event loop
        """
        consumer = self.uninstantiated_things[id]
        instance = consumer(**kwargs, eventloop_name=self.instance_name) # type: Thing
        self.things.append(instance)
        rpc_server = instance.rpc_server
        self.request_listener_loop.call_soon(asyncio.create_task(lambda : rpc_server.poll()))
        self.request_listener_loop.call_soon(asyncio.create_task(lambda : rpc_server.tunnel_message_to_things()))
        if not self.threaded:
            self.thing_executor_loop.call_soon(asyncio.create_task(lambda : self.run_single_target(instance)))
        else: 
            _thing_executor = threading.Thread(target=self.run_things_executor, args=([instance],))
            _thing_executor.start()

    def run(self):
        """
        start the eventloop
        """
        if not self.threaded:
            _thing_executor = threading.Thread(target=self.run_things_executor, args=(self.things,))
            _thing_executor.start()
        else: 
            for thing in self.things:
                _thing_executor = threading.Thread(target=self.run_things_executor, args=([thing],))
                _thing_executor.start()
        self.run_external_message_listener()
        if not self.threaded:
            _thing_executor.join()


    @classmethod
    def get_async_loop(cls):
        """
        get or automatically create an asnyc loop for the current thread.
        """
        try:
            loop = asyncio.get_event_loop()
        except RuntimeError:
            loop = asyncio.new_event_loop()
            # set_event_loop_policy() - why not?
            asyncio.set_event_loop(loop)
        return loop
        

    def run_external_message_listener(self):
        """
        Runs ZMQ's sockets which are visible to clients.
        This method is automatically called by ``run()`` method. 
        Please dont call this method when the async loop is already running. 
        """
        self.request_listener_loop = self.get_async_loop()
        rpc_servers = [thing.rpc_server for thing in self.things]
        futures = []
        for rpc_server in rpc_servers:
            futures.append(rpc_server.poll())
            futures.append(rpc_server.tunnel_message_to_things())
        self.logger.info("starting external message listener thread")
        self.request_listener_loop.run_until_complete(asyncio.gather(*futures))
        pending_tasks = asyncio.all_tasks(self.request_listener_loop)
        self.request_listener_loop.run_until_complete(asyncio.gather(*pending_tasks))
        self.logger.info("exiting external listener event loop {}".format(self.instance_name))
        self.request_listener_loop.close()
    

    def run_things_executor(self, things):
        """
        Run ZMQ sockets which provide queued instructions to ``Thing``.
        This method is automatically called by ``run()`` method. 
        Please dont call this method when the async loop is already running. 
        """
        thing_executor_loop = self.get_async_loop()
        self.thing_executor_loop = thing_executor_loop # atomic assignment for thread safety
        self.logger.info(f"starting thing executor loop in thread {threading.get_ident()} for {[obj.instance_name for obj in things]}")
        thing_executor_loop.run_until_complete(
            asyncio.gather(*[self.run_single_target(instance) for instance in things])
        )
        self.logger.info(f"exiting event loop in thread {threading.get_ident()}")
        thing_executor_loop.close()


    @classmethod
    async def run_single_target(cls, instance : Thing) -> None: 
        instance_name = instance.instance_name
        while True:
            instructions = await instance.message_broker.async_recv_instructions()
            for instruction in instructions:
                client, _, client_type, _, msg_id, _, instruction_str, arguments, context = instruction
                oneway = context.pop('oneway', False)
                fetch_execution_logs = context.pop("fetch_execution_logs", False)
                if fetch_execution_logs:
                    list_handler = ListHandler([])
                    list_handler.setLevel(logging.DEBUG)
                    list_handler.setFormatter(instance.logger.handlers[0].formatter)
                    instance.logger.addHandler(list_handler)
                try:
                    instance.logger.debug(f"client {client} of client type {client_type} issued instruction " +
                                f"{instruction_str} with message id {msg_id}. starting execution.")
                    return_value = await cls.execute_once(instance_name, instance, instruction_str, arguments) #type: ignore 
                    if oneway:
                        await instance.message_broker.async_send_reply_with_message_type(instruction, b'ONEWAY', None)
                        continue
                    if fetch_execution_logs:
                        return_value = {
                            "returnValue" : return_value,
                            "execution_logs" : list_handler.log_list
                        }
                    await instance.message_broker.async_send_reply(instruction, return_value)
                    # Also catches exception in sending messages like serialization error
                except (BreakInnerLoop, BreakAllLoops):
                    instance.logger.info("Thing {} with instance name {} exiting event loop.".format(
                                                            instance.__class__.__name__, instance_name))
                    if oneway:
                        await instance.message_broker.async_send_reply_with_message_type(instruction, b'ONEWAY', None)
                        continue
                    return_value = None
                    if fetch_execution_logs:
                        return_value = { 
                            "returnValue" : None,
                            "execution_logs" : list_handler.log_list
                        }
                    await instance.message_broker.async_send_reply(instruction, return_value)
                    return 
                except Exception as ex:
                    instance.logger.error("Thing {} with instance name {} produced error : {}.".format(
                                                            instance.__class__.__name__, instance_name, ex))
                    if oneway:
                        await instance.message_broker.async_send_reply_with_message_type(instruction, b'ONEWAY', None)
                        continue
                    return_value = dict(exception= format_exception_as_json(ex))
                    if fetch_execution_logs:
                        return_value["execution_logs"] = list_handler.log_list
                    await instance.message_broker.async_send_reply_with_message_type(instruction, 
                                                                    b'EXCEPTION', return_value)
                finally:
                    if fetch_execution_logs:
                        instance.logger.removeHandler(list_handler)

    @classmethod
    async def execute_once(cls, instance_name : str, instance : Thing, instruction_str : str, 
                           arguments : typing.Dict[str, typing.Any]) -> typing.Dict[str, typing.Any]:
        resource = instance.instance_resources.get(instruction_str, None) 
        if resource is None:
            raise AttributeError(f"unknown remote resource represented by instruction {instruction_str}")
        if resource.isaction:      
            if resource.state is None or (hasattr(instance, 'state_machine') and 
                            instance.state_machine.current_state in resource.state):
                # Note that because we actually find the resource within __prepare_instance__, its already bound
                # and we dont have to separately bind it. 
                if resource.schema_validator is not None:
                    resource.schema_validator.validate(arguments)
                
                func = resource.obj
                args = arguments.pop('__args__', tuple())
                if resource.iscoroutine:
                    if resource.isparameterized:
                        if len(args) > 0:
                            raise RuntimeError("parameterized functions cannot have positional arguments")
                        return await func(resource.bound_obj, *args, **arguments)
                    return await func(*args, **arguments) # arguments then become kwargs
                else:
                    if resource.isparameterized:
                        if len(args) > 0:
                            raise RuntimeError("parameterized functions cannot have positional arguments")
                        return func(resource.bound_obj, *args, **arguments)
                    return func(*args, **arguments) # arguments then become kwargs
            else: 
                raise StateMachineError("Thing '{}' is in '{}' state, however command can be executed only in '{}' state".format(
                        instance_name, instance.state, resource.state))
        
        elif resource.isproperty:
            action = instruction_str.split('/')[-1]
            prop = resource.obj # type: Property
            owner_inst = resource.bound_obj # type: Thing
            if action == "write": 
                if resource.state is None or (hasattr(instance, 'state_machine') and  
                                        instance.state_machine.current_state in resource.state):
                    if isinstance(arguments, dict) and len(arguments) == 1 and 'value' in arguments:
                        return prop.__set__(owner_inst, arguments['value'])
                    return prop.__set__(owner_inst, arguments)
                else: 
                    raise StateMachineError("Thing {} is in `{}` state, however attribute can be written only in `{}` state".format(
                        instance_name, instance.state_machine.current_state, resource.state))
            elif action == "read":
                return prop.__get__(owner_inst, type(owner_inst))             
            elif action == "delete":
                if prop.fdel is not None:
                    return prop.fdel() # this may not be correct yet
                raise NotImplementedError("This property does not support deletion")
        raise NotImplementedError("Unimplemented execution path for Thing {} for instruction {}".format(instance_name, instruction_str))


def fork_empty_eventloop(instance_name : str, logfile : typing.Union[str, None] = None, python_command : str = 'python',
                        condaenv : typing.Union[str, None] = None, prefix_command : typing.Union[str, None] = None):
    command_str = '{}{}{}-c "from hololinked.server import EventLoop; E = EventLoop({}); E.run();"'.format(
        f'{prefix_command} ' if prefix_command is not None else '',
        f'call conda activate {condaenv} && ' if condaenv is not None else '',
        f'{python_command} ',
        f"instance_name = '{instance_name}', logfile = '{logfile}'"
    )
    print(f"command to invoke : {command_str}")
    subprocess.Popen(
        command_str, 
        shell = True
    )


# class ForkedEventLoop:

#     def __init__(self, instance_name : str, things : Union[Thing, Consumer, List[Union[Thing, Consumer]]], 
#                 log_level : int = logging.INFO, **kwargs):
#         self.subprocess = Process(target = forked_eventloop, kwargs = dict(
#                         instance_name = instance_name, 
#                         things = things, 
#                         log_level = log_level,
#                         **kwargs
#                     ))
    
#     def start(self):
#         self.Process.start()



__all__ = ['EventLoop', 'Consumer', 'fork_empty_eventloop']